extern crate bitvec;
extern crate hercules_ir;

use std::collections::{BTreeMap, VecDeque};
use std::fmt::{Error, Write};
use std::iter::{zip, FromIterator};

use self::bitvec::prelude::*;

use self::hercules_ir::*;

use crate::*;

/*
 * Entry Hercules functions are lowered to async Rust code to achieve easy task
 * level parallelism. This Rust is generated textually, and is included via a
 * procedural macro in the user's Rust code.
 */
pub fn rt_codegen<W: Write>(
    func_id: FunctionID,
    module: &Module,
    reverse_postorder: &Vec<NodeID>,
    typing: &Vec<TypeID>,
    control_subgraph: &Subgraph,
    antideps: &Vec<(NodeID, NodeID)>,
    bbs: &Vec<NodeID>,
    collection_objects: &CollectionObjects,
    callgraph: &CallGraph,
    devices: &Vec<Device>,
    w: &mut W,
) -> Result<(), Error> {
    let ctx = RTContext {
        func_id,
        module,
        reverse_postorder,
        typing,
        control_subgraph,
        antideps,
        bbs,
        collection_objects,
        callgraph,
        devices,
    };
    ctx.codegen_function(w)
}

struct RTContext<'a> {
    func_id: FunctionID,
    module: &'a Module,
    reverse_postorder: &'a Vec<NodeID>,
    typing: &'a Vec<TypeID>,
    control_subgraph: &'a Subgraph,
    antideps: &'a Vec<(NodeID, NodeID)>,
    bbs: &'a Vec<NodeID>,
    collection_objects: &'a CollectionObjects,
    callgraph: &'a CallGraph,
    devices: &'a Vec<Device>,
}

impl<'a> RTContext<'a> {
    fn codegen_function<W: Write>(&self, w: &mut W) -> Result<(), Error> {
        let func = &self.get_func();

        // Dump the function signature.
        write!(
            w,
            "#[allow(unused_variables,unused_mut)]\nasync fn {}(",
            func.name
        )?;
        let mut first_param = true;
        // The first set of parameters are dynamic constants.
        for idx in 0..func.num_dynamic_constants {
            if first_param {
                first_param = false;
            } else {
                write!(w, ", ")?;
            }
            write!(w, "dc_p{}: u64", idx)?;
        }
        // The second set of parameters are normal parameters.
        for idx in 0..func.param_types.len() {
            if first_param {
                first_param = false;
            } else {
                write!(w, ", ")?;
            }
            if !self.module.types[func.param_types[idx].idx()].is_primitive() {
                write!(w, "mut ")?;
            }
            write!(
                w,
                "p_i{}: {}",
                idx,
                self.get_type_interface(func.param_types[idx])
            )?;
        }
        write!(w, ") -> {} {{\n", self.get_type_interface(func.return_type))?;

        // Copy the "interface" parameters to "non-interface" parameters.
        // The purpose of this is to convert collection objects from a Box<[u8]>
        // type to a *mut u8 type. This name copying is done so that we can
        // easily construct objects just after this by moving the "inferface"
        // parameters.
        for (idx, ty) in func.param_types.iter().enumerate() {
            if self.module.types[ty.idx()].is_primitive() {
                write!(w, "    let p{} = p_i{};\n", idx, idx)?;
            } else {
                write!(
                    w,
                    "    let p{} = ::std::boxed::Box::as_mut_ptr(&mut p_i{}) as *mut u8;\n",
                    idx, idx
                )?;
            }
        }

        // Collect the boxes representing ownership over collection objects for
        // this function. The actual emitted computation is done entirely using
        // pointers, so these get emitted to hold onto ownership over the
        // underlying memory and to automatically clean them up when this
        // function returns. Collection objects are inside Options, since their
        // ownership may get passed to other called RT functions. If this
        // function returns a collection object, then at the very end, right
        // before the return, the to-be-returned pointer is compared against the
        // owned collection objects - it should match exactly one of those
        // objects, and that box is what's actually returned.
        let mem_obj_ty = "::core::option::Option<::std::boxed::Box<[u8]>>";
        for object in self.collection_objects[&self.func_id].iter_objects() {
            match self.collection_objects[&self.func_id].origin(object) {
                CollectionObjectOrigin::Parameter(index) => write!(
                    w,
                    "    let mut obj{}: {} = Some(p_i{});\n",
                    object.idx(),
                    mem_obj_ty,
                    index
                )?,
                CollectionObjectOrigin::Constant(id) => {
                    let size = self.codegen_type_size(self.typing[id.idx()]);
                    write!(
                        w,
                        "    let mut obj{}: {} = Some((0..{}).map(|_| 0u8).collect());\n",
                        object.idx(),
                        mem_obj_ty,
                        size
                    )?
                }
                CollectionObjectOrigin::Call(_) | CollectionObjectOrigin::Undef(_) => write!(
                    w,
                    "    let mut obj{}: {} = None;\n",
                    object.idx(),
                    mem_obj_ty,
                )?,
            }
        }

        // Dump signatures for called CPU functions.
        write!(w, "    extern \"C\" {{\n")?;
        for callee in self.callgraph.get_callees(self.func_id) {
            if self.devices[callee.idx()] != Device::LLVM {
                continue;
            }
            let callee = &self.module.functions[callee.idx()];
            write!(w, "        fn {}(", callee.name)?;
            let mut first_param = true;
            for idx in 0..callee.num_dynamic_constants {
                if first_param {
                    first_param = false;
                } else {
                    write!(w, ", ")?;
                }
                write!(w, "dc{}: u64", idx)?;
            }
            for (idx, ty) in callee.param_types.iter().enumerate() {
                if first_param {
                    first_param = false;
                } else {
                    write!(w, ", ")?;
                }
                write!(w, "p{}: {}", idx, self.get_type(*ty))?;
            }
            write!(w, ") -> {};\n", self.get_type(callee.return_type))?;
        }
        write!(w, "    }}\n")?;

        // Declare intermediary variables for every value.
        for idx in 0..func.nodes.len() {
            if func.nodes[idx].is_control() {
                continue;
            }
            write!(
                w,
                "    let mut node_{}: {} = {};\n",
                idx,
                self.get_type(self.typing[idx]),
                if self.module.types[self.typing[idx].idx()].is_integer() {
                    "0"
                } else if self.module.types[self.typing[idx].idx()].is_float() {
                    "0.0"
                } else {
                    "::core::ptr::null::<u8>() as _"
                }
            )?;
        }

        // The core executor is a Rust loop. We literally run a "control token"
        // as described in the original sea of nodes paper through the basic
        // blocks to drive execution.
        write!(
            w,
            "    let mut control_token: i8 = 0;\n    loop {{\n        match control_token {{\n",
        )?;

        let mut blocks: BTreeMap<_, _> = (0..func.nodes.len())
            .filter(|idx| func.nodes[*idx].is_control())
            .map(|idx| (NodeID::new(idx), String::new()))
            .collect();

        // Emit data flow into basic blocks.
        let mut worklist = VecDeque::from_iter(
            self.reverse_postorder
                .into_iter()
                .filter(|id| !func.nodes[id.idx()].is_control()),
        );
        let mut visited = bitvec![u8, Lsb0; 0; func.nodes.len()];
        let antideps = flip_antideps(&self.antideps);
        while let Some(id) = worklist.pop_front() {
            let node = &func.nodes[id.idx()];
            if node.is_phi()
                || node.is_reduce()
                || get_uses(node)
                    .as_ref()
                    .into_iter()
                    .chain(antideps.get(&id).into_iter().flatten())
                    .all(|u| {
                        func.nodes[u.idx()].is_control()
                            || self.bbs[u.idx()] != self.bbs[id.idx()]
                            || visited[u.idx()]
                    })
            {
                self.codegen_data_node(*id, &mut blocks)?;
                visited.set(id.idx(), true);
            } else {
                worklist.push_back(id);
            }
        }

        // Emit control flow into basic blocks.
        for id in (0..func.nodes.len()).map(NodeID::new) {
            if !func.nodes[id.idx()].is_control() {
                continue;
            }
            self.codegen_control_node(id, &mut blocks)?;
        }

        // Dump the emitted basic blocks.
        for (id, block) in blocks {
            write!(
                w,
                "            {} => {{\n{}            }}\n",
                id.idx(),
                block
            )?;
        }

        // Close the match, loop, and function.
        write!(w, "            _ => panic!()\n        }}\n    }}\n}}\n")?;
        Ok(())
    }

    /*
     * While control nodes in Hercules IR are predecessor-centric (each take a
     * control input that defines the predecessor relationship), the Rust loop
     * we generate is successor centric. This difference requires explicit
     * translation.
     */
    fn codegen_control_node(
        &self,
        id: NodeID,
        blocks: &mut BTreeMap<NodeID, String>,
    ) -> Result<(), Error> {
        let func = &self.get_func();
        match func.nodes[id.idx()] {
            // Start, region, and projection control nodes all have exactly one
            // successor and are otherwise simple.
            Node::Start
            | Node::Region { preds: _ }
            | Node::Projection {
                control: _,
                selection: _,
            } => {
                let block = &mut blocks.get_mut(&id).unwrap();
                let succ = self.control_subgraph.succs(id).next().unwrap();
                write!(block, "                control_token = {};\n", succ.idx())?
            }
            // If nodes have two successors - examine the projections to
            // determine which branch is which, and branch between them.
            Node::If { control: _, cond } => {
                let block = &mut blocks.get_mut(&id).unwrap();
                let mut succs = self.control_subgraph.succs(id);
                let succ1 = succs.next().unwrap();
                let succ2 = succs.next().unwrap();
                let succ1_is_true = func.nodes[succ1.idx()].try_projection(1).is_some();
                write!(
                    block,
                    "                control_token = if {} {{ {} }} else {{ {} }};\n",
                    self.get_value(cond),
                    if succ1_is_true { succ1 } else { succ2 }.idx(),
                    if succ1_is_true { succ2 } else { succ1 }.idx(),
                )?
            }
            Node::Return { control: _, data } => {
                let block = &mut blocks.get_mut(&id).unwrap();
                let objects = self.collection_objects[&self.func_id].objects(data);
                if objects.is_empty() {
                    write!(block, "                return {};\n", self.get_value(data))?
                } else {
                    // If the value to return is a collection object, figure out
                    // which object it actually is at runtime and return that
                    // box.
                    for object in objects {
                        write!(block, "                if let Some(mut obj) = obj{} && ::std::boxed::Box::as_mut_ptr(&mut obj) as *mut u8 == {} {{\n", object.idx(), self.get_value(data))?;
                        write!(block, "                    return obj;\n")?;
                        write!(block, "                }}\n")?;
                    }
                    write!(block, "                panic!(\"HERCULES PANIC: Pointer to be returned doesn't match any known collection objects.\");\n")?
                }
            }
            _ => panic!("PANIC: Can't lower {:?}.", func.nodes[id.idx()]),
        }
        Ok(())
    }

    /*
     * Lower data nodes in Hercules IR into Rust statements.
     */
    fn codegen_data_node(
        &self,
        id: NodeID,
        blocks: &mut BTreeMap<NodeID, String>,
    ) -> Result<(), Error> {
        let func = &self.get_func();
        match func.nodes[id.idx()] {
            Node::Parameter { index } => {
                let block = &mut blocks.get_mut(&self.bbs[id.idx()]).unwrap();
                write!(
                    block,
                    "                {} = p{};\n",
                    self.get_value(id),
                    index
                )?
            }
            Node::Constant { id: cons_id } => {
                let block = &mut blocks.get_mut(&self.bbs[id.idx()]).unwrap();
                write!(block, "                {} = ", self.get_value(id))?;
                match self.module.constants[cons_id.idx()] {
                    Constant::Boolean(val) => write!(block, "{}bool", val)?,
                    Constant::Integer8(val) => write!(block, "{}i8", val)?,
                    Constant::Integer16(val) => write!(block, "{}i16", val)?,
                    Constant::Integer32(val) => write!(block, "{}i32", val)?,
                    Constant::Integer64(val) => write!(block, "{}i64", val)?,
                    Constant::UnsignedInteger8(val) => write!(block, "{}u8", val)?,
                    Constant::UnsignedInteger16(val) => write!(block, "{}u16", val)?,
                    Constant::UnsignedInteger32(val) => write!(block, "{}u32", val)?,
                    Constant::UnsignedInteger64(val) => write!(block, "{}u64", val)?,
                    Constant::Float32(val) => write!(block, "{}f32", val)?,
                    Constant::Float64(val) => write!(block, "{}f64", val)?,
                    Constant::Product(_, _) | Constant::Summation(_, _, _) | Constant::Array(_) => {
                        let objects = self.collection_objects[&self.func_id].objects(id);
                        assert_eq!(objects.len(), 1);
                        let object = objects[0];
                        write!(
                            block,
                            "::std::boxed::Box::as_mut_ptr(obj{}.as_mut().unwrap()) as *mut u8",
                            object.idx()
                        )?
                    }
                }
                write!(block, ";\n")?
            }
            Node::Call {
                control: _,
                function: callee_id,
                ref dynamic_constants,
                ref args,
            } => {
                match self.devices[callee_id.idx()] {
                    Device::LLVM => {
                        let block = &mut blocks.get_mut(&self.bbs[id.idx()]).unwrap();
                        write!(
                            block,
                            "                {} = unsafe {{ {}(",
                            self.get_value(id),
                            self.module.functions[callee_id.idx()].name
                        )?;
                        for dc in dynamic_constants {
                            self.codegen_dynamic_constant(*dc, block)?;
                            write!(block, ", ")?;
                        }
                        for arg in args {
                            write!(block, "{}, ", self.get_value(*arg))?;
                        }
                        write!(block, ") }};\n")?;

                        // When a CPU function is called that returns a
                        // collection object, that object must have come from
                        // one of its parameters. Dynamically figure out which
                        // one it came from, so that we can move it to the slot
                        // of the output object.
                        let call_objects = self.collection_objects[&self.func_id].objects(id);
                        if !call_objects.is_empty() {
                            assert_eq!(call_objects.len(), 1);
                            let call_object = call_objects[0];

                            let callee_returned_objects =
                                self.collection_objects[&callee_id].returned_objects();
                            let possible_params: Vec<_> =
                                (0..self.module.functions[callee_id.idx()].param_types.len())
                                    .filter(|idx| {
                                        let object_of_param = self.collection_objects[&callee_id]
                                            .param_to_object(*idx);
                                        // Look at parameters that could be the
                                        // source of the memory object returned
                                        // by the function.
                                        object_of_param
                                            .map(|object_of_param| {
                                                callee_returned_objects.contains(&object_of_param)
                                            })
                                            .unwrap_or(false)
                                    })
                                    .collect();
                            let arg_objects = args
                                .into_iter()
                                .enumerate()
                                .filter(|(idx, _)| possible_params.contains(idx))
                                .map(|(_, arg)| {
                                    self.collection_objects[&self.func_id]
                                        .objects(*arg)
                                        .into_iter()
                                })
                                .flatten();

                            // Dynamically check which of the memory objects
                            // corresponding to arguments to the call was
                            // returned by the call. Move that memory object
                            // into the memory object of the call.
                            let mut first_obj = true;
                            for arg_object in arg_objects {
                                write!(block, "                ")?;
                                if first_obj {
                                    first_obj = false;
                                } else {
                                    write!(block, "else ")?;
                                }
                                write!(block, "if let Some(obj) = obj{}.as_mut() && ::std::boxed::Box::as_mut_ptr(obj) as *mut u8 == {} {{\n", arg_object.idx(), self.get_value(id))?;
                                write!(
                                    block,
                                    "                    obj{} = obj{}.take();\n",
                                    call_object.idx(),
                                    arg_object.idx()
                                )?;
                                write!(block, "                }}\n")?;
                            }
                            write!(block, "                else {{\n")?;
                            write!(block, "                    panic!(\"HERCULES PANIC: Pointer returned from called function doesn't match any known collection objects.\");\n")?;
                            write!(block, "                }}\n")?;
                        }
                    }
                    Device::AsyncRust => {
                        let block = &mut blocks.get_mut(&self.bbs[id.idx()]).unwrap();
                        write!(
                            block,
                            "                {} = {}(",
                            self.get_value(id),
                            self.module.functions[callee_id.idx()].name
                        )?;
                        for dc in dynamic_constants {
                            self.codegen_dynamic_constant(*dc, block)?;
                            write!(block, ", ")?;
                        }
                        for arg in args {
                            if self.module.types[self.typing[arg.idx()].idx()].is_primitive() {
                                write!(block, "{}, ", self.get_value(*arg))?;
                            } else {
                                write!(block, "{}.take(), ", self.get_value(*arg))?;
                            }
                        }
                        write!(block, ").await;\n")?;
                    }
                    _ => todo!(),
                }
            }
            _ => panic!(
                "PANIC: Can't lower {:?} in {}.",
                func.nodes[id.idx()],
                func.name
            ),
        }
        Ok(())
    }

    /*
     * Lower dynamic constant in Hercules IR into a Rust expression.
     */
    fn codegen_dynamic_constant<W: Write>(
        &self,
        id: DynamicConstantID,
        w: &mut W,
    ) -> Result<(), Error> {
        match self.module.dynamic_constants[id.idx()] {
            DynamicConstant::Constant(val) => write!(w, "{}", val)?,
            DynamicConstant::Parameter(idx) => write!(w, "dc_p{}", idx)?,
            DynamicConstant::Add(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(left, w)?;
                write!(w, "+")?;
                self.codegen_dynamic_constant(right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Sub(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(left, w)?;
                write!(w, "-")?;
                self.codegen_dynamic_constant(right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Mul(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(left, w)?;
                write!(w, "*")?;
                self.codegen_dynamic_constant(right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Div(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(left, w)?;
                write!(w, "/")?;
                self.codegen_dynamic_constant(right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Rem(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(left, w)?;
                write!(w, "%")?;
                self.codegen_dynamic_constant(right, w)?;
                write!(w, ")")?;
            }
        }
        Ok(())
    }

    /*
     * Lower the size of a type into a Rust expression.
     */
    fn codegen_type_size(&self, ty: TypeID) -> String {
        match self.module.types[ty.idx()] {
            Type::Control => panic!(),
            Type::Boolean | Type::Integer8 | Type::UnsignedInteger8 => "1".to_string(),
            Type::Integer16 | Type::UnsignedInteger16 => "2".to_string(),
            Type::Integer32 | Type::UnsignedInteger32 | Type::Float32 => "4".to_string(),
            Type::Integer64 | Type::UnsignedInteger64 | Type::Float64 => "8".to_string(),
            Type::Product(ref fields) => {
                let fields_align = fields
                    .into_iter()
                    .map(|id| get_type_alignment(&self.module.types, *id));
                let fields: Vec<String> = fields
                    .into_iter()
                    .map(|id| self.codegen_type_size(*id))
                    .collect();

                // Emit LLVM IR to round up to the alignment of the next field,
                // and then add the size of that field. At the end, round up to
                // the alignment of the whole struct.
                let mut acc_size = "0".to_string();
                for (field_align, field) in zip(fields_align, fields) {
                    acc_size = format!(
                        "(({} + {}) & !{})",
                        acc_size,
                        field_align - 1,
                        field_align - 1
                    );
                    acc_size = format!("({} + {})", acc_size, field);
                }
                let total_align = get_type_alignment(&self.module.types, ty);
                format!(
                    "(({} + {}) & !{})",
                    acc_size,
                    total_align - 1,
                    total_align - 1
                )
            }
            Type::Summation(ref variants) => {
                let variants = variants.into_iter().map(|id| self.codegen_type_size(*id));

                // The size of a summation is the size of the largest field,
                // plus 1 byte and alignment for the discriminant.
                let mut acc_size = "0".to_string();
                for variant in variants {
                    acc_size = format!("::core::cmp::max({}, {})", acc_size, variant);
                }

                // No alignment is necessary for the 1 byte discriminant.
                let total_align = get_type_alignment(&self.module.types, ty);
                format!(
                    "(({} + 1 + {}) & !{})",
                    acc_size,
                    total_align - 1,
                    total_align - 1
                )
            }
            Type::Array(elem, ref bounds) => {
                // The size of an array is the size of the element multipled by
                // the dynamic constant bounds.
                let mut acc_size = self.codegen_type_size(elem);
                for dc in bounds {
                    acc_size = format!("{} * ", acc_size);
                    self.codegen_dynamic_constant(*dc, &mut acc_size).unwrap();
                }
                format!("({})", acc_size)
            }
        }
    }

    fn get_func(&self) -> &Function {
        &self.module.functions[self.func_id.idx()]
    }

    fn get_value(&self, id: NodeID) -> String {
        format!("node_{}", id.idx())
    }

    fn get_type(&self, id: TypeID) -> &'static str {
        convert_type(&self.module.types[id.idx()])
    }

    fn get_type_interface(&self, id: TypeID) -> &'static str {
        convert_type_interface(&self.module.types[id.idx()])
    }
}

fn convert_type(ty: &Type) -> &'static str {
    match ty {
        Type::Boolean => "bool",
        Type::Integer8 => "i8",
        Type::Integer16 => "i16",
        Type::Integer32 => "i32",
        Type::Integer64 => "i64",
        Type::UnsignedInteger8 => "u8",
        Type::UnsignedInteger16 => "u16",
        Type::UnsignedInteger32 => "u32",
        Type::UnsignedInteger64 => "u64",
        Type::Float32 => "f32",
        Type::Float64 => "f64",
        Type::Product(_) | Type::Summation(_) | Type::Array(_, _) => "*mut u8",
        _ => panic!(),
    }
}

/*
 * Collection types are passed to / returned from runtime functions through a
 * wrapper type for ownership tracking reasons.
 */
fn convert_type_interface(ty: &Type) -> &'static str {
    match ty {
        Type::Product(_) | Type::Summation(_) | Type::Array(_, _) => "Box<[u8]>",
        _ => convert_type(ty),
    }
}
