use std::collections::{BTreeMap, BTreeSet};
use std::iter::zip;

use crate::*;

/*
 * In addition to collections, we need to figure out which "generation" of a
 * collection a node may take as input.
 */
#[derive(PartialEq, Eq, Clone, Debug)]
struct GenerationLattice {
    objs: BTreeSet<(CollectionObjectID, NodeID)>,
}

impl Semilattice for GenerationLattice {
    fn meet(a: &Self, b: &Self) -> Self {
        GenerationLattice {
            objs: a.objs.union(&b.objs).map(|x| *x).collect(),
        }
    }

    fn top() -> Self {
        GenerationLattice {
            objs: BTreeSet::new(),
        }
    }

    fn bottom() -> Self {
        // Bottom is not representable for this lattice with our Semilattice
        // interface, but we never need to construct it.
        panic!()
    }
}

/*
 * Function to assemble anti-dependence edges. Returns a list of pairs of nodes.
 * The first item in the pair is the reading node, and the second item is the
 * mutating node.
 */
pub fn antideps(
    function: &Function,
    reverse_postorder: &Vec<NodeID>,
    objects: &FunctionCollectionObjects,
) -> Vec<(NodeID, NodeID)> {
    // First, we analyze "generations" of collections as they are mutated.
    // Originating nodes and mutating nodes start a new generation of a
    // collection. Generations are not ordered due to loops, but are rather just
    // node IDs of the originating or mutating node (parameter, constant, call,
    // undef, write). Other nodes operating on collections mean reads / writes
    // can operate on potentially different generations of multiple collections
    // (phi, reduce, select).
    let lattice = forward_dataflow(function, reverse_postorder, |inputs, id| {
        match function.nodes[id.idx()] {
            Node::Phi {
                control: _,
                data: _,
            }
            | Node::Reduce {
                control: _,
                init: _,
                reduct: _,
            }
            | Node::Ternary {
                op: TernaryOperator::Select,
                first: _,
                second: _,
                third: _,
            } => inputs
                .into_iter()
                .fold(GenerationLattice::top(), |acc, input| {
                    GenerationLattice::meet(&acc, input)
                }),
            Node::Parameter { index: _ } | Node::Constant { id: _ } | Node::Undef { ty: _ } => {
                let objs = objects.objects(id);
                GenerationLattice {
                    objs: objs.into_iter().map(|obj| (*obj, id)).collect(),
                }
            }
            Node::Call {
                control: _,
                function: _,
                dynamic_constants: _,
                ref args,
            } => {
                let mut objs = BTreeSet::new();
                let call_objs = objects.objects(id);

                // If this call node might originate an object, add that to the
                // lattice output - its generation is this call node.
                for obj in call_objs {
                    if objects.origin(*obj) == CollectionObjectOrigin::Call(id) {
                        assert!(objs.len() <= 1);
                        objs.insert((*obj, id));
                    }
                }

                // For every argument...
                for (arg, arg_gens) in zip(args, inputs.into_iter().skip(1)) {
                    // Look at its objects...
                    for arg_obj in objects.objects(*arg) {
                        // For each object that might be returned...
                        if call_objs.contains(&arg_obj) {
                            let mutable = objects.mutators(*arg_obj).contains(&id);
                            for (obj, gen) in arg_gens.objs.iter() {
                                // Add that object to the output lattice.
                                if obj == arg_obj && mutable {
                                    // Set the generation to this node if the
                                    // object might be mutated.
                                    objs.insert((*obj, id));
                                } else if obj == arg_obj {
                                    // Otherwise, keep the old generation.
                                    objs.insert((*obj, *gen));
                                }
                            }
                        }
                    }
                }
                GenerationLattice { objs }
            }
            Node::Read {
                collect: _,
                indices: _,
            } => inputs[0].clone(),
            Node::Write {
                collect: _,
                data: _,
                indices: _,
            } => {
                // Writes update the generation to the write.
                let objs = inputs[0].objs.iter().map(|(obj, _)| (*obj, id)).collect();
                GenerationLattice { objs }
            }
            _ => GenerationLattice::top(),
        }
    });

    // Second, we generate anti-dependence edges from the dataflow analysis.
    // There are three cases where an anti-dependence edge is generated:
    //
    // 1. A read node and a write node share an object and generation pair on
    //    their `collect` input.
    // 2. A read node and a call node share an object and generation pair, where
    //    the pair is on the read's `collect` input and the pair is on any input
    //    of the call node AND the call node is a mutator of the object.
    // 3. A call node and a write node share an object and generation pair,
    //    where the pair is on any input of the call node and the pair is on the
    //    write's `collect` input.
    let mut reads_writes_calls_mut_calls_per_pair: BTreeMap<
        (CollectionObjectID, NodeID),
        (Vec<NodeID>, Vec<NodeID>, Vec<NodeID>, Vec<NodeID>),
    > = BTreeMap::new();
    for (idx, node) in function.nodes.iter().enumerate() {
        let id = NodeID::new(idx);
        match node {
            Node::Read {
                collect,
                indices: _,
            } => {
                for pair in lattice[collect.idx()].objs.iter() {
                    reads_writes_calls_mut_calls_per_pair
                        .entry(*pair)
                        .or_default()
                        .0
                        .push(id);
                }
            }
            Node::Write {
                collect,
                data: _,
                indices: _,
            } => {
                for pair in lattice[collect.idx()].objs.iter() {
                    reads_writes_calls_mut_calls_per_pair
                        .entry(*pair)
                        .or_default()
                        .1
                        .push(id);
                }
            }
            Node::Call {
                control: _,
                function: _,
                dynamic_constants: _,
                ref args,
            } => {
                for arg in args {
                    for pair in lattice[arg.idx()].objs.iter() {
                        if objects.mutators(pair.0).contains(&id) {
                            reads_writes_calls_mut_calls_per_pair
                                .entry(*pair)
                                .or_default()
                                .3
                                .push(id);
                        } else {
                            reads_writes_calls_mut_calls_per_pair
                                .entry(*pair)
                                .or_default()
                                .2
                                .push(id);
                        }
                    }
                }
            }
            _ => {}
        }
    }

    // Once we've matched reads / writes / calls by object and generation pair,
    // the pair itself no longer matters.
    let mut antideps = vec![];
    for (_, (reads, writes, calls, mut_calls)) in reads_writes_calls_mut_calls_per_pair {
        // Case 1:
        for read in reads.iter() {
            for write in writes.iter() {
                antideps.push((*read, *write));
            }
        }

        // Case 2:
        for read in reads.iter() {
            for mut_call in mut_calls.iter() {
                antideps.push((*read, *mut_call));
            }
        }

        // Case 3:
        for call in calls.iter().chain(mut_calls.iter()) {
            for write in writes.iter() {
                antideps.push((*call, *write));
            }
        }
    }

    antideps
}

/*
 * Utility to make a map from node to anti-dependency uses (map mutator ->
 * reads).
 */
pub fn flip_antideps(antideps: &Vec<(NodeID, NodeID)>) -> BTreeMap<NodeID, Vec<NodeID>> {
    let mut result: BTreeMap<NodeID, Vec<NodeID>> = BTreeMap::new();

    for (read, mutator) in antideps {
        result.entry(*mutator).or_default().push(*read);
    }

    result
}
