extern crate hercules_ir;
use std::collections::HashMap;

use self::hercules_ir::ir::*;
use crate::*;
use std::iter::zip;

/**
 * Given an editor for each function in a module, return V s.t.
 * V[i] = true iff every call node to the function with index i
 * is editable. If there are no calls to this function, V[i] = true.
 */
fn get_editable_callsites(editors: &mut Vec<FunctionEditor>) -> Vec<bool> {
    let mut callsites_editable = vec![true; editors.len()];
    for editor in editors {
        for (idx, (_, function, _, _)) in editor
            .func()
            .nodes
            .iter()
            .enumerate()
            .filter_map(|(idx, node)| node.try_call().map(|c| (idx, c)))
        {
            if !editor.is_mutable(NodeID::new(idx)) {
                callsites_editable[function.idx()] = false;
            }
        }
    }
    callsites_editable
}

/**
 * Given a type tree, return a Vec containing all leaves which are not units.
 */
fn get_nonempty_leaves(edit: &FunctionEdit, type_id: &TypeID) -> Vec<TypeID> {
    let ty = edit.get_type(*type_id).clone();
    match ty {
        Type::Product(type_ids) => {
            let mut leaves = vec![];
            for type_id in type_ids {
                leaves.extend(get_nonempty_leaves(&edit, &type_id))
            }
            leaves
        }
        _ => vec![*type_id],
    }
}

/**
 * Given a `source` NodeID which produces a product containing
 * all nonempty leaves of the type tree for `type_id` in order, build
 * a node producing the `type_id`.
 *
 * `offset` represents the index at which to begin reading
 * elements of the `source` product.
 *
 * Returns a 3-tuple of
 * 1. Node producing the `type`
 * 2. "Next" offset, i.e. `offset` + number of reads performed to build (1)
 * 3. List of node IDs which read `source` (tracked so that these will not
 *    be replaced by replace_all_uses_where)
 */
fn build_uncompressed_product(
    edit: &mut FunctionEdit,
    source: &NodeID,
    type_id: &TypeID,
    offset: usize,
) -> (NodeID, usize, Vec<NodeID>) {
    let ty = edit.get_type(*type_id).clone();
    match ty {
        Type::Product(child_type_ids) => {
            // Step 1. Create an empty constant for the type. We'll write
            // child values into this constant.
            let empty_constant_id = edit.add_zero_constant(*type_id);
            let empty_constant_node = edit.add_node(Node::Constant {
                id: empty_constant_id,
            });
            // Step 2. Build a node that generates each inner type.
            // Since `source` contains nonempty leaves *in order*,
            // we must process inner types in order; as part of this,
            // inner type i+1 must read from where inner type i left off,
            // hence we track the `current_offset` at which we are reading.
            // Similarly, to combine results of all recursive calls,
            // we keep the invariant that, at iteration i+1, currently_writing_to
            // is an instance of `type_id` for which the first i elements
            // have been populated based on inorder nonempty leaves
            // (and, at iteration 0, it is empty).
            let mut current_offset = offset;
            let mut currently_writing_to = empty_constant_node;
            let mut readers = vec![];
            for (idx, child_type_id) in child_type_ids.iter().enumerate() {
                let (child_data, next_offset, child_readers) =
                    build_uncompressed_product(edit, source, child_type_id, current_offset);
                current_offset = next_offset;
                currently_writing_to = edit.add_node(Node::Write {
                    collect: currently_writing_to,
                    data: child_data,
                    indices: Box::new([Index::Field(idx)]),
                });
                readers.extend(child_readers)
            }
            (currently_writing_to, current_offset, readers)
        }
        _ => {
            // If the type is not a product, then we've reached a nonempty
            // leaf, which we must read from source. Since this is a single
            // read, the new offset increases by only 1.
            let reader = edit.add_node(Node::Read {
                collect: *source,
                indices: Box::new([Index::Field(offset)]),
            });
            (reader, offset + 1, vec![reader])
        }
    }
}

/**
 * Given a node with a product value, read the product's values
 * *in order* into the nonempty leaves of a product type represented
 * by type_id. Returns the ID of the resulting node, as well as the IDs
 * of all nodes which read from `node_id`.
 */
fn uncompress_product(
    edit: &mut FunctionEdit,
    node_id: &NodeID,
    type_id: &TypeID,
) -> (NodeID, Vec<NodeID>) {
    let (uncompressed_value, _, readers) = build_uncompressed_product(edit, node_id, type_id, 0);
    (uncompressed_value, readers)
}

/**
* Let `read_from` be a node with a value of type `type_id`.
* Let `source` be a product value.
* Returns a node representing the value obtained by writing
* nonempty leaves of `read_from` *in order* into `source`,
* starting at `offset`.
*
* `source` should be a product type with at least enough indices
* to support this operation. Typically, `build_compressed_product`
* should be called initially with a `source` created by adding a
* zero constant for the flattened `type_id`.
*
* Returns:
* 1. The ID of the node to which all nonempty leaves have been written
* 2. The first offset after `offset` which was not written to.
*/
fn build_compressed_product(
    mut edit: &mut FunctionEdit,
    source: &NodeID,
    type_id: &TypeID,
    offset: usize,
    read_from: &NodeID,
) -> (NodeID, usize) {
    let ty = edit.get_type(*type_id).clone();
    match ty {
        Type::Product(child_type_ids) => {
            // Iterate through child types in order. For each type, construct
            // a node that reads the corresponding value from `read_from`,
            // and pass it as the node to read from in the recursive call.
            let mut next_offset = offset;
            let mut next_destination = *source;
            for (idx, child_type_id) in child_type_ids.iter().enumerate() {
                let child_value = edit.add_node(Node::Read {
                    collect: *read_from,
                    indices: Box::new([Index::Field(idx)]),
                });
                (next_destination, next_offset) = build_compressed_product(
                    &mut edit,
                    &next_destination,
                    &child_type_id,
                    next_offset,
                    &child_value,
                );
            }
            (next_destination, next_offset)
        }
        _ => {
            let writer = edit.add_node(Node::Write {
                collect: *source,
                data: *read_from,
                indices: Box::new([Index::Field(offset)]),
            });
            (writer, offset + 1)
        }
    }
}

/**
 * Given a node which has a value of the given type (which must be a product)
 * generate a new product node created by inserting nonempty leaves of the
 * source node *in order*. Returns the ID of this node, as well as the ID of
 * its type.
 */
fn compress_product(
    edit: &mut FunctionEdit,
    node_id: &NodeID,
    type_id: &TypeID,
) -> (NodeID, TypeID) {
    let nonempty_leaves = get_nonempty_leaves(&edit, &type_id);
    let compressed_type = Type::Product(nonempty_leaves.into_boxed_slice());
    let compressed_type_id = edit.add_type(compressed_type);

    let empty_compressed_constant_id = edit.add_zero_constant(compressed_type_id);
    let empty_compressed_node_id = edit.add_node(Node::Constant {
        id: empty_compressed_constant_id,
    });

    let (compressed_value, _) =
        build_compressed_product(edit, &empty_compressed_node_id, type_id, 0, node_id);

    (compressed_value, compressed_type_id)
}

fn compress_return_products(editors: &mut Vec<FunctionEditor>, all_callsites_editable: &Vec<bool>) {
    // Track whether we successfully applied edits to return statements,
    // so that callsites are only modified when returns were. This is
    // initialized to false, so that `is_compressed` is false when
    // the corresponding entry in `callsites_editable` is false.
    let mut is_compressed = vec![false; editors.len()];
    let old_return_type_ids: Vec<_> = editors
        .iter()
        .map(|editor| editor.func().return_type)
        .collect();

    // Step 1. Track mapping of dynamic constant indexes to ids, so that
    // we can substitute when generating empty constants later. The reason
    // this works is that the following property is satisfied:
    //   Let f and g be two functions such that f has d_f dynamic constants
    //   and g has d_g dynamic constants. Wlog assume d_f < d_g. Then, the
    //   first d_f dynamic constants of g are the dynamic constants of f.
    // For any call node, the ith dynamic constant in the node is provided
    // for the ith dynamic constant of the function called. So, when we need
    // to take a type and replace d function dynamic constants with their
    // values from a call, it suffices to look at the first d entries of
    // dc_param_idx_to_dc_id to get the id of the dynamic constants in the function,
    // and then replace dc_param_idx_to_dc_id[i] with call.dynamic_constants[i],
    // for all i.
    let mut found_idxs = HashMap::new();
    for id in editors[0].dynamic_constant_ids() {
        let dc = editors[0].get_dynamic_constant(id);
        if let DynamicConstant::Parameter(idx) = *dc {
            assert!(!found_idxs.contains_key(&idx));
            found_idxs.insert(idx, id);
        }
    }
    let mut dc_param_idx_to_dc_id = vec![];
    for idx in 0..found_idxs.len() {
        dc_param_idx_to_dc_id.push(found_idxs[&idx]);
    }

    // Step 2. Modify the return type of all editors corresponding to a function
    // for which we can edit every callsite, and the return type is a product.
    for (idx, editor) in editors.iter_mut().enumerate() {
        if !all_callsites_editable[idx] {
            continue;
        }

        let old_return_id = NodeID::new(
            (0..editor.func().nodes.len())
                .filter(|idx| editor.func().nodes[*idx].is_return())
                .next()
                .unwrap(),
        );
        let old_return_type_id = old_return_type_ids[idx];

        is_compressed[idx] = editor.get_type(editor.func().return_type).is_product()
            && editor.edit(|mut edit| {
                let return_node = edit.get_node(old_return_id);
                let (return_control, return_data) = return_node.try_return().unwrap();

                let (compressed_data_id, compressed_type_id) =
                    compress_product(&mut edit, &return_data, &old_return_type_id);

                edit.set_return_type(compressed_type_id);
                let new_return_id = edit.add_node(Node::Return {
                    control: return_control,
                    data: compressed_data_id,
                });
                edit.sub_edit(old_return_id, new_return_id);
                let edit = edit.replace_all_uses(old_return_id, new_return_id)?;
                edit.delete_node(old_return_id)
            });
    }

    // Step 3: For every editor, update all mutable callsites corresponding to
    // calls to functions which have been compressed. Since we only compress returns
    // for functions for which every callsite is mutable, this should never fail,
    // so we panic if it does.
    for (_, editor) in editors.iter_mut().enumerate() {
        let call_node_ids: Vec<_> = (0..editor.func().nodes.len())
            .map(NodeID::new)
            .filter(|id| editor.func().nodes[id.idx()].is_call())
            .filter(|id| editor.is_mutable(*id))
            .collect();

        for call_node_id in call_node_ids {
            let (_, function_id, ref dynamic_constants, _) =
                editor.func().nodes[call_node_id.idx()].try_call().unwrap();
            if !is_compressed[function_id.idx()] {
                continue;
            }

            // Before creating the uncompressed product, we must update
            // the type of the uncompressed product to reflect the dynamic
            // constants provided when calling the function. Since we can
            // only replace one constant at a time, we need to map
            // constants to dummy values, and then map these to the
            // replacement values (this prevents the case of replacements
            // (0->1), (1->2) causing conflicts when we have [0, 1], we should
            // get [1, 2], not [2, 2], which a naive loop would generate).

            // A similar loop exists in the inline pass but at the node level.
            // If this becomes a common pattern, it would be worth creating
            // a better abstraction around bulk replacement.

            let new_dcs = (*dynamic_constants).clone();

            let edit_successful = editor.edit(|mut edit| {
                let old_dcs = dc_param_idx_to_dc_id[..new_dcs.len()].to_vec().clone();
                let mut substituted = old_return_type_ids[function_id.idx()];

                let first_dc = edit.num_dynamic_constants() + 1;
                for (dc_a, dc_n) in zip(old_dcs, first_dc..) {
                    substituted = substitute_dynamic_constants_in_type(
                        dc_a,
                        DynamicConstantID::new(dc_n),
                        substituted,
                        &mut edit,
                    );
                }

                for (dc_n, dc_b) in zip(first_dc.., new_dcs.iter()) {
                    substituted = substitute_dynamic_constants_in_type(
                        DynamicConstantID::new(dc_n),
                        *dc_b,
                        substituted,
                        &mut edit,
                    );
                }

                let (expanded_product, readers) =
                    uncompress_product(&mut edit, &call_node_id, &substituted);
                edit.replace_all_uses_where(call_node_id, expanded_product, |id| {
                    !readers.contains(id)
                })
            });

            if !edit_successful {
                panic!("Tried and failed to edit mutable callsite!");
            }
        }
    }
}

fn remove_return_singletons(editors: &mut Vec<FunctionEditor>, all_callsites_editable: &Vec<bool>) {
    // Track whether we removed a singleton product from the return of each
    // editor's function. Defaults to false so that if the function was not
    // edited (i.e. because not all callsites are editable), then no callsites
    // will be edited.
    let mut singleton_removed = vec![false; editors.len()];
    let old_return_type_ids: Vec<_> = editors
        .iter()
        .map(|editor| editor.func().return_type)
        .collect();

    // Step 1. For all editors which correspond to a function for whic hall
    // callsites are editable, modify their return type by extracting the
    // value from the singleton and returning it directly.
    for (idx, editor) in editors.iter_mut().enumerate() {
        if !all_callsites_editable[idx] {
            continue;
        }

        let return_type = editor.get_type(old_return_type_ids[idx]).clone();
        singleton_removed[idx] = match return_type {
            Type::Product(tys) if tys.len() == 1 && all_callsites_editable[idx] => {
                let old_return_id = NodeID::new(
                    (0..editor.func().nodes.len())
                        .filter(|idx| editor.func().nodes[*idx].is_return())
                        .next()
                        .unwrap(),
                );

                editor.edit(|mut edit| {
                    let (old_control, old_data) =
                        edit.get_node(old_return_id).try_return().unwrap();

                    let extracted_singleton_id = edit.add_node(Node::Read {
                        collect: old_data,
                        indices: Box::new([Index::Field(0)]),
                    });
                    let new_return_id = edit.add_node(Node::Return {
                        control: old_control,
                        data: extracted_singleton_id,
                    });
                    edit.sub_edit(old_return_id, new_return_id);
                    edit.set_return_type(tys[0]);

                    edit.delete_node(old_return_id)
                })
            }
            _ => false,
        }
    }

    // Step 2. For each editor, find all callsites and reconstruct
    // the singleton product at each if the return of the corresponding
    // function was modified. This should always succeed since we only
    // edited functions for which all callsites were mutable, so panic
    // if an edit does not succeed.
    for editor in editors.iter_mut() {
        let call_node_ids: Vec<_> = (0..editor.func().nodes.len())
            .map(NodeID::new)
            .filter(|id| editor.func().nodes[id.idx()].is_call())
            .filter(|id| editor.is_mutable(*id))
            .collect();

        for call_node_id in call_node_ids {
            let (_, function, _, _) = editor.func().nodes[call_node_id.idx()].try_call().unwrap();

            if singleton_removed[function.idx()] {
                let edit_successful = editor.edit(|mut edit| {
                    let empty_constant_id =
                        edit.add_zero_constant(old_return_type_ids[function.idx()]);
                    let empty_node_id = edit.add_node(Node::Constant {
                        id: empty_constant_id,
                    });

                    let restored_singleton_id = edit.add_node(Node::Write {
                        collect: empty_node_id,
                        data: call_node_id,
                        indices: Box::new([Index::Field(0)]),
                    });
                    edit.replace_all_uses_where(call_node_id, restored_singleton_id, |id| {
                        *id != restored_singleton_id
                    })
                });

                if !edit_successful {
                    panic!("Tried and failed to edit mutable callsite!");
                }
            }
        }
    }
}

pub fn interprocedural_sroa(editors: &mut Vec<FunctionEditor>) {
    // SROA is implemented in two phases. First, we flatten (or "compress")
    // all product return types, so that they are only depth 1 products,
    // and do not contain any empty products.
    // Next, if any return type is now a singleton product, we
    // remove the singleton and just retun the type directly.
    // We only apply these changes to functions for which
    // all their callsites are editable.
    let all_callsites_editable = get_editable_callsites(editors);
    compress_return_products(editors, &all_callsites_editable);
    remove_return_singletons(editors, &all_callsites_editable);

    // Run DCE to prevent issues with schedule repair.
    for editor in editors.iter_mut() {
        dce(editor);
    }
}
