extern crate itertools;
extern crate ordered_float;

use crate::value;

use self::itertools::Itertools;
use std::collections::HashMap;
use std::panic;
use std::collections::hash_map::Entry::Occupied;

use value::*;

extern crate hercules_ir;
extern crate hercules_opt;

use self::hercules_ir::*;

/* High level design details / discussion for this:
 *
 * This crate includes tools for interpreting a hercules IR module. Execution model / flow is based on
 * "A Simple Graph-Based Intermediate Representation" Section 3 by Cliff Click.
 *
 * In the future we would an *interactive* interpreter. I.e to be able to step forward and backwards
 * throughout the IR (idealy visually) and analyze results at certain steps in the computation.
 */
pub struct FunctionExecutionState<'a> {
    args: Vec<InterpreterVal>, // parameters

    // Map a vec of (thread indicies, reduce_node) -> Value
    // Reduction values are shared across the fork / join pair that drives them, so they can't be local to a control token.
    reduce_values: HashMap<(Vec<usize>, NodeID), InterpreterVal>, // Map (thread indicies, reduction node) -> value
    join_counters: HashMap<(Vec<usize>, NodeID), usize>, // Map (thread indicies, join_node) -> join_counter. Counters until the joins are done!
    dynamic_constant_params: Vec<usize>,

    // Per module State.
    module: &'a Module,
    function_contexts: &'a Vec<FunctionContext<'a>>,

    function_id: FunctionID,
}

pub struct FunctionContext<'a> {
    control_subgraph: &'a Subgraph,
    def_use: &'a ImmutableDefUseMap,
    fork_join_map: &'a HashMap<NodeID, NodeID>, // Map forks -> joins
    fork_join_nest: &'a HashMap<NodeID, Vec<NodeID>>,
}

impl <'a> FunctionContext<'a> {
    pub fn new  (
        control_subgraph: &'a Subgraph,
        def_use: &'a ImmutableDefUseMap,
        fork_join_map: &'a HashMap<NodeID, NodeID>, // Map forks -> joins
        fork_join_nest: &'a HashMap<NodeID, Vec<NodeID>>,
    ) -> FunctionContext<'a> {
        FunctionContext {
            control_subgraph,
            def_use,
            fork_join_map,
            fork_join_nest,
        }
    }
}

pub fn dyn_const_value(dc: &DynamicConstant, dyn_const_params: &[usize]) -> usize {
    match dc {
        DynamicConstant::Constant(v) => *v,
        DynamicConstant::Parameter(v) => dyn_const_params[*v],
    }
}
// Each control token stores a current position, and also a mapping of fork nodes -> thread idx.
#[derive(Debug, Clone, Eq, PartialEq)]
pub struct ControlToken {
    pub curr: NodeID,
    pub prev: NodeID, // Used for properly latching PHI nodes, as they depend on their control predecessor.
    pub thread_indicies: Vec<usize>, // Stack of thread_indicies, can use fork_join nests to recover thread indicies per fork.

    // Map phis -> values
    // Oddly, these are stored per control token, when some other per-token data is stored in the FunctionExecutionState
    // we could move this `phi_values` to be consistent, however that would make CoW behvaior optimization more difficult.
    pub phi_values: HashMap<NodeID, InterpreterVal>, // TODO: Optimize this to a Cow<HashMap>, or optimize this whole thing.
}

impl ControlToken {
    pub fn moved_to(&self, next: NodeID) -> ControlToken {
        ControlToken { curr: next, prev: self.curr, thread_indicies: self.thread_indicies.clone(), phi_values: self.phi_values.clone() }
    }
}
impl<'a> FunctionExecutionState<'a> {
    pub fn new(
        args: Vec<InterpreterVal>,
        module: &'a Module,
        function_id: FunctionID,
        function_contexts: &'a Vec<FunctionContext>,
        dynamic_constant_params: Vec<usize>,
    ) -> Self {
        assert_eq!(args.len(), module.functions[function_id.idx()].param_types.len());

        FunctionExecutionState {
            args,
            reduce_values: HashMap::new(),
            join_counters: HashMap::new(),
            module,
            function_contexts,
            function_id,
            dynamic_constant_params,
        }
    }

    pub fn get_control_subgraph(&self) -> &Subgraph {
        self.function_contexts[self.function_id.idx()].control_subgraph
    }

    pub fn get_fork_join_map(&self) -> &HashMap<NodeID, NodeID> {
        self.function_contexts[self.function_id.idx()].fork_join_map
    }

    pub fn get_fork_join_nest(&self) -> &HashMap<NodeID, Vec<NodeID>> {
        self.function_contexts[self.function_id.idx()].fork_join_nest
    }

    pub fn get_def_use(&self) -> &ImmutableDefUseMap {
        self.function_contexts[self.function_id.idx()].def_use
    }

    pub fn get_function(&self) -> &Function {
        &self.module.functions[self.function_id.idx()]
    }

    /* Drives PHI values of this region for a control token, returns the next control node. */
    pub fn handle_region(
        &mut self,
        token: &mut ControlToken,
        preds: &Box<[NodeID]>,
    ) -> NodeID {

        let prev = token.prev;
        let node = token.curr;
        
        // Gather PHI nodes for this region node.
        let phis: Vec<NodeID> = self
            .get_def_use()
            .get_users(node)
            .into_iter()
            .filter_map(|n| {
                if self.get_function().nodes[n.idx()].is_phi() {
                    Some(*n)
                } else {
                    None
                }
            })
            .collect();

        // Find incoming edge of this region.
        let (edge_num, _) = preds
            .into_iter()
            .find_position(|pred| **pred == prev)
            .expect("PANIC: invalid incoming edge.");

        let mut to_latch = vec![];

        // Get values to latch at this point.
        for phi in phis {
            to_latch.push(self.handle_phi(token, edge_num, phi));
        }

        // Write values into map.
        for (phi, value) in to_latch {
            token.phi_values.insert(phi, value);
        }

        // Return the (single) control successor of the region node
        self.get_control_subgraph().succs(node).next().unwrap()
    }

    pub fn handle_phi(
        &mut self,
        token: &ControlToken,
        edge: usize,
        phi: NodeID,
    ) -> (NodeID, InterpreterVal) {
        let (_control, data) = &self.get_function().nodes[phi.idx()]
            .try_phi()
            .expect("PANIC: handle_phi on non-phi node.");
        let value_node = data[edge];
        // println!("Latching PHI value of node {:?}", value_node.idx());
        let value = self.handle_data(token, value_node);
        (phi, value)
    }

    // Drive all reductions for this join.
    // TODO (@xrouth): Change all these to return vecs of ControlTokens, so this is consistent with other handle_* functions.
    pub fn handle_join(&mut self, mut token: ControlToken, join: NodeID) -> Option<ControlToken> {
        let reduces: Vec<NodeID> = self
            .get_def_use()
            .get_users(join)
            .into_iter()
            .filter_map(|n| {
                if self.get_function().nodes[n.idx()].is_reduce() {
                    Some(*n)
                } else {
                    None
                }
            })
            .collect();

        for reduction in reduces {
            self.handle_reduction(&token, reduction);
        }

        let thread_values = self.get_thread_factors(&token, join);
        // dbg!(thread_values.clone());
        // This and_modify doesn't do aynthing??
        self.join_counters
            .entry((thread_values.clone(), join))
            .and_modify(|v| *v -= 1);

        if *self
            .join_counters
            .get(&(thread_values, join))
            .expect("PANIC: join counter not initialized")
            == 0
        {
            let curr = token.curr;
            token.prev = curr;
            token.thread_indicies.pop(); // Get rid of this thread index.
            token.curr = self.get_control_subgraph().succs(join).next().unwrap();
            Some(token)
        } else {
            None
        }
    }

    /* Get top N-1 thread factors, where N is the depth of the control token. */
    pub fn get_thread_factors(&self, token: &ControlToken, control_node: NodeID) -> Vec<usize> {
        let nested_forks = self
            .get_fork_join_nest()
            .get(&control_node)
            .expect("PANIC: invalid node for fork join nest.")
            .clone();
        // Take the top N entries such that it matches the length of the TRF in the control token.

        // Get the depth of the control token that is requesting this reduction node.
        let fork_levels = nested_forks.len();

        let len = fork_levels - 1;
        let mut thread_values = token.thread_indicies.clone();
        thread_values.truncate(len);
        thread_values
    }

    pub fn intialize_reduction(&mut self, token_at_fork: &ControlToken, reduce: NodeID) {

        let token = token_at_fork;

        let (control, init, _) = &self.get_function().nodes[reduce.idx()]
            .try_reduce()
            .expect("PANIC: handle_reduction on non-reduce node.");

        let thread_values = self.get_thread_factors(token, *control);

        let init = self.handle_data(&token, *init);
        // Q (@xrouth): It is UB to have the initializer depend on things within the fork-join section? do we check for that?
        // A: Should be done in verify (TODO).

        self.reduce_values
            .entry((thread_values.clone(), reduce))
            .or_insert(init);
    }

    // Drive the reduction, this will be invoked for each control token.
    pub fn handle_reduction(&mut self, token: &ControlToken, reduce: NodeID) {
        let (control, _, reduct) = &self.get_function().nodes[reduce.idx()]
            .try_reduce()
            .expect("PANIC: handle_reduction on non-reduce node.");

        let thread_values = self.get_thread_factors(token, *control);

        // If empty set to default (figure out how to not repeat this check)
        // TODO: (Can we do it upon entry to the fork node?) (YES!)

        let data = self.handle_data(&token, *reduct);
        /* 
        println!(
            "reduction write: {:?}, {:?}, {:?}",
            thread_values, reduce, data
        ); */

        self.reduce_values.insert((thread_values, reduce), data);
    }

    pub fn handle_data(&mut self, token: &ControlToken, node: NodeID) -> InterpreterVal {
        // println!("Data Node: {} {:?}", node.idx(), &self.get_function().nodes[node.idx()]);

        // Partial borrow complaint. :/
        match &self.module.functions[self.function_id.idx()].nodes[node.idx()]{
            Node::Phi { control: _, data: _ } => (*token
                .phi_values
                .get(&node)
                .expect("PANIC: Phi value not latched."))
            .clone(),
            Node::ThreadID { control } => {
                // `control` is the fork that drives this node.
                let nesting_level = self
                    .get_fork_join_nest()
                    .get(control)
                    .expect("PANIC: No nesting information for thread index!")
                    .len();
                let v = token.thread_indicies[nesting_level - 1]; // Might have to -1?
                InterpreterVal::DynamicConstant((v).into())
            }
            // If we read from a reduction that is the same depth as this thread, we need to write back to it before anyone else reads from it.
            // This probably isn't the exact condition, but somethign similar. Anyways, we achieve correctness by iterating control nodes recursively.
            Node::Reduce {
                control,
                init,
                reduct: _,
            } => {
                let thread_values = self.get_thread_factors(token, *control);

                // println!("reduction read: {:?}, {:?}", thread_values, node);
                let entry = self
                    .reduce_values
                    .entry((thread_values, node));
                
                let val = match entry {
                    Occupied(v) => v.get().clone(),
                    std::collections::hash_map::Entry::Vacant(_) => panic!("Reduce has not been initialized!"),
                };
                // println!("value: {:?}", val.clone());
                val
            }
            Node::Parameter { index } => self.args[*index].clone(),
            Node::Constant { id } => {
                let con = &self.module.constants[id.idx()];
                InterpreterVal::from_constant(
                    con,
                    &self.module.constants,
                    &self.module.types,
                    &self.module.dynamic_constants,
                    &self.dynamic_constant_params
                )
            }
            Node::DynamicConstant { id } => {
                let dyn_con = &self.module.dynamic_constants[id.idx()];
                let v = match dyn_con {
                    DynamicConstant::Constant(v) => v,
                    DynamicConstant::Parameter(v) => &self.dynamic_constant_params[*v],
                };
                // TODO: Figure out what type / semantics are of thread ID and dynamic const.
                InterpreterVal::DynamicConstant((*v).into())
            }
            Node::Unary { input, op } => {
                let val = self.handle_data(token, *input);
                InterpreterVal::unary_op(*op, val)
            }
            Node::Binary { left, right, op } => {
                let left = self.handle_data(token, *left);
                let right = self.handle_data(token, *right);

                InterpreterVal::binary_op(*op, left, right)
            }
            Node::Ternary {
                first,
                second,
                third,
                op,
            } => {
                let first = self.handle_data(token, *first);
                let second = self.handle_data(token, *second);
                let third = self.handle_data(token, *third);

                match op {
                    TernaryOperator::Select => {
                        if first == InterpreterVal::Boolean(true) {
                            second
                        } else {
                            third
                        }
                    }
                }
            }
            Node::Call {
                function,
                dynamic_constants,
                args,
            } => {

                let args = args.into_iter()
                            .map(|arg_node| self.handle_data(token, *arg_node))
                            .collect();


                let dynamic_constant_params = dynamic_constants.into_iter()
                            .map(|id| {
                                let dyn_con = &self.module.dynamic_constants[id.idx()];
                                let v = match dyn_con {
                                    DynamicConstant::Constant(v) => *v,
                                    DynamicConstant::Parameter(v) => self.dynamic_constant_params[*v],
                                };
                                v
                            }).collect_vec();

                let mut state = FunctionExecutionState::new(
                    args,
                    self.module,
                    *function,
                    self.function_contexts,
                    dynamic_constant_params,
                );

                state.run()
            }
            Node::Read { collect, indices } => {
                let collection = self.handle_data(token, *collect);

                self.handle_read(token, collection, indices)
            }
            Node::Write {
                collect,
                data,
                indices,
            } => {
                let collection = self.handle_data(token, *collect);
                let data = self.handle_data(token, *data);
                self.handle_write(token, collection, data, indices)
            }
            _ => todo!(),
        }
    }

    pub fn handle_write(
        &mut self,
        token: &ControlToken,
        collection: InterpreterVal,
        data: InterpreterVal,
        indices: &[Index],
    ) -> InterpreterVal {
        let index = &indices[0];

        // TODO (@xrouth): Recurse on writes correctly
        let val = match index {
            Index::Field(_) => todo!(),
            Index::Variant(_) => todo!(),
            Index::Position(array_indices) => {
                // Arrays also have inner indices...
                // Recover dimensional data from types.
                let array_indices: Vec<_> = array_indices
                    .into_iter()
                    .map(|idx| self.handle_data(token, *idx).as_usize())
                    .collect();
                // TODO: Implemenet . try_array() and other try_conversions on the InterpreterVal type
                if let InterpreterVal::Array(type_id, mut vals) = collection {
                    // TODO: Make this its own funciton to reuse w/ array_size
                    let extents: Vec<_> = self.module.types[type_id.idx()]
                        .try_extents()
                        .expect("PANIC: wrong type for array")
                        .into_iter()
                        .map(|extent| dyn_const_value(&self.module.dynamic_constants[extent.idx()], &self.dynamic_constant_params))
                        .collect();
                    let idx = InterpreterVal::array_idx(&extents, &array_indices);
                    //println!("idx: {:?}", idx);
                    vals[idx] = data;
                    InterpreterVal::Array(type_id, vals)
                } else {
                    panic!("PANIC: Position index on not an array")
                }
            }
        };
        val
    }

    pub fn handle_read(
        &mut self,
        token: &ControlToken,
        collection: InterpreterVal,
        indices: &[Index],
    ) -> InterpreterVal {
        let index = &indices[0];

        let val = match index {
            Index::Field(_) => todo!(),
            Index::Variant(_) => todo!(),
            Index::Position(array_indices) => {
                // Arrays also have inner indices...
                // Recover dimensional data from types.
                let array_indices: Vec<_> = array_indices
                    .into_iter()
                    .map(|idx| self.handle_data(token, *idx).as_usize())
                    .collect();

                // TODO: Implemenet . try_array() and other try_conversions on the InterpreterVal type
                if let InterpreterVal::Array(type_id, vals) = collection {
                    // TODO: Make this its own funciton to reuse w/ array_size
                    let extents: Vec<_> = self.module.types[type_id.idx()]
                        .try_extents()
                        .expect("PANIC: wrong type for array")
                        .into_iter()
                        .map(|extent| dyn_const_value(&self.module.dynamic_constants[extent.idx()], &self.dynamic_constant_params))
                        .collect();
                    vals[InterpreterVal::array_idx(&extents, &array_indices)].clone()
                } else {
                    panic!("PANIC: Position index on not an array")
                }
            }
        };

        // Recurse
        if indices.len() == 1 {
            val
        } else {
            self.handle_read(token, val, &indices[1..])
        }
    }

    pub fn run(&mut self) -> InterpreterVal {
        let start_node: NodeID = NodeID::new(0);

        let start_token = ControlToken {
            curr: start_node,
            prev: start_node,
            thread_indicies: Vec::new(),
            phi_values: HashMap::new(),
        };

        let mut live_tokens: Vec<ControlToken> = Vec::new();
        live_tokens.push(start_token);


        // To do reduction nodes correctly we have to traverse control tokens in a depth-first fashion (i.e immediately handle spawned threads).
        'outer: loop {
            let mut ctrl_token = live_tokens.pop().expect("PANIC: Interpreter ran out of control tokens without returning.");

            /* println!(
                "\n\nNew Token at: Control State: {} threads: {:?}, {:?}",
                ctrl_token.curr.idx(),
                ctrl_token.thread_indicies.clone(),
                &self.get_function().nodes[ctrl_token.curr.idx()]
            ); */

            // TODO: Rust is annoying and can't recognize that this is a partial borrow.
            // Can't partial borrow, so need a clone. 
            let node = &self.get_function().nodes[ctrl_token.curr.idx()].clone();
            let new_tokens = match node {
                Node::Start => {
                    let next: NodeID = self.get_control_subgraph().succs(ctrl_token.curr).next().unwrap();

                    let ctrl_token = ctrl_token.moved_to(next);
    
                    vec![ctrl_token]
                }
                Node::Region { preds } => {

                    // Updates 
                    let next = self.handle_region(&mut ctrl_token, &preds);
                    let ctrl_token = ctrl_token.moved_to(next);

                    vec![ctrl_token]
                }
                Node::If { control: _, cond } => {
                    let cond = self.handle_data(&ctrl_token, *cond);

                    // Convert condition to usize
                    let cond: usize = match cond {
                        InterpreterVal::Boolean(v) => v.into(),
                        _ => panic!("PANIC: Invalid condition for IF, please typecheck."),
                    };

                    // Find control read succesor that matches the condition.
                    let next = self
                        .get_control_subgraph()
                        .succs(ctrl_token.curr)
                        .find(|n| {
                            self.get_function().nodes[n.idx()]
                                .try_projection(cond)
                                .is_some()
                        })
                        .expect("PANIC: No outgoing valid outgoing edge.");

                    let ctrl_token = ctrl_token.moved_to(next);
                    vec![ctrl_token]
                }
                Node::Projection { .. } => {
                    let next: NodeID = self.get_control_subgraph().succs(ctrl_token.curr).next().unwrap();

                    let ctrl_token = ctrl_token.moved_to(next);

                    vec![ctrl_token]
                }

                Node::Match { control: _, sum: _ } => todo!(),
                Node::Fork { control: _, factor } => {
                    let fork = ctrl_token.curr;
                    let dyn_con = &self.module.dynamic_constants[factor.idx()];

                    let thread_factor = match dyn_con {
                        DynamicConstant::Constant(v) => v,
                        DynamicConstant::Parameter(v) => &self.dynamic_constant_params[*v],
                    }.clone();

                    // Update control token 
                    let next = self.get_control_subgraph().succs(ctrl_token.curr).nth(0).unwrap();
                    let ctrl_token = ctrl_token.moved_to(next);

                    let mut tokens_to_add = Vec::with_capacity(thread_factor);

                    assert_ne!(thread_factor, 0);

                    // Token is at its correct sontrol succesor already.
                    // Add the new thread index.
                    for i in 0..(thread_factor) {
                        let mut new_token = ctrl_token.clone(); // Copy map, curr, prev, etc.
                        new_token.thread_indicies.push(i); // Stack of thread indicies
                        tokens_to_add.push(new_token);
                    }

                    let thread_factors = self.get_thread_factors(&ctrl_token, ctrl_token.curr);

                    // Find join and initialize them, and set their reduction counters as well.
                    let join = self
                        .get_fork_join_map()
                        .get(&fork)
                        .expect("PANIC: fork missing a join.")
                        .clone();

                    let reduces: Vec<NodeID> = self
                        .get_def_use()
                        .get_users(join)
                        .into_iter()
                        .filter_map(|n| {
                            if self.get_function().nodes[n.idx()].is_reduce() {
                                Some(*n)
                            } else {
                                None
                            }
                        })
                        .collect();
        
                    for reduction in reduces {
                        // TODO: Is this the correct reduction?
                        self.intialize_reduction(&ctrl_token, reduction);
                    }


                    self.join_counters.insert((thread_factors, join), thread_factor);

                    tokens_to_add
                }
                Node::Join { control: _ } => {
                    let join = ctrl_token.curr;

                    // Only make a control token if the join is finished, checked in handle_join.
                    if let Some(next) = self.handle_join(ctrl_token, join) {
                        vec![next]
                    } else {
                        vec![]
                    }
                }
                Node::Return { control: _, data } => {
                    let result = self.handle_data(&ctrl_token, *data);
                    // println!("result = {:?}", result);
                    break 'outer result;
                }
                _ => {
                    panic!("PANIC: Unexpected node in control subgraph {:?}", node);
                }
            };

            for i in new_tokens {
                live_tokens.push(i);
            }

        }
    }
}


